# -*- coding: utf-8 -*-
import gc
import io
import logging
import os
import pathlib
import sys
from typing import Union

import mincepy
import pytest  # pylint: disable=wrong-import-order

import minkipy
from test import script as script_module  # pylint: disable=wrong-import-order

# pylint: disable=unused-argument, invalid-name, no-member


def test_load_script():
    script = """\
def add(a, b):
    return a + b
"""
    script_file = io.StringIO(script)
    module = minkipy.load_script(script_file)
    assert module.add(5, 6) == 11

    # Test loading by file path
    module = minkipy.load_script(minkipy.constants.FILE_PREFIX + script_module.__file__)
    assert module.add(5, 6) == 11

    # Test loading by import
    module = minkipy.load_script(minkipy.constants.MODULE_PREFIX + script_module.__name__)
    assert module.add(5, 6) == 11

    with pytest.raises(TypeError):
        minkipy.load_script(5)


def writing_task(path, msg: str):
    """A simple task that writes a message to a file"""
    with open(str(path), 'w') as file:
        file.write(msg)


def my_task(arg):
    return arg


def test_create_task(tmp_path, test_project):
    # Create a task from a function in a script
    with minkipy.utils.working_directory(tmp_path):
        task = minkipy.task(my_task, [5])
        assert isinstance(task, minkipy.Task)
        assert task.run() == 5
        assert task.state == minkipy.DONE

        task2 = minkipy.task('{}@my_task'.format(__file__), [10])
        assert task2.run() == 10
        assert task.state == minkipy.DONE


def test_task_working_folder(tmp_path, test_project):
    """Check that a script runs in the folder given"""
    folder = 'task_folder'
    msg = 'Do you have a license for this minki?'
    task = minkipy.task(writing_task, ['hello.txt', msg], folder=folder)

    with minkipy.utils.working_directory(tmp_path):
        task.run()
    assert os.path.exists(str(tmp_path / folder / 'hello.txt'))
    with open(str(tmp_path / folder / 'hello.txt'), 'r') as file:
        assert file.read() == msg


def test_script_written(tmp_path, test_project):
    """Test that a task writes it's script to the folder"""
    folder = 'task_folder'
    task = minkipy.task(my_task, [10], folder=folder)

    with minkipy.utils.working_directory(tmp_path):
        task.run()
    assert os.path.isfile(str(tmp_path / folder / pathlib.Path(__file__).name))


def do_some_logging(level):
    logging.log(level, "I've got some bad news...")


def test_task_logging(tmp_path, test_project):
    task = minkipy.task(do_some_logging, (logging.DEBUG,), folder='test_task')
    task.log_level = logging.DEBUG - 1
    task.save()

    with minkipy.utils.working_directory(tmp_path):
        task.run()

    log_text = task.log_file.read_text()
    assert "I've got some bad news" in log_text
    assert str(task.obj_id) in log_text

    # Now turn logging off
    task = minkipy.task(do_some_logging, (logging.WARNING,), folder='test_task')
    task.log_level = None
    with minkipy.utils.working_directory(tmp_path):
        task.run()
    assert "I've got some bad news" not in task.log_file.read_text()


def show_msg(msg, err=False):
    print(msg, file=sys.stdout if not err else sys.stderr)


def test_task_stds(tmp_path, test_project):
    """Test that tasks can capture standard out and err"""
    with minkipy.utils.working_directory(tmp_path):
        task = minkipy.task(show_msg, args=('Hello stdout!',), folder='test_task')
        task.run()
        assert 'Hello stdout!' in task.stdout.read_text()
        task_id = task.save()

        task2 = minkipy.task(show_msg, args=('Hello stderr!', True), folder='test_task')
        task2.run()
        assert 'Hello stderr!' in task2.stderr.read_text()
        task2_id = task2.save()

        # Make sure they are really stored when saving
        del task, task2
        loaded, loaded2 = mincepy.load(task_id, task2_id)
        assert 'Hello stdout!' in loaded.stdout.read_text()
        assert 'Hello stderr!' in loaded2.stderr.read_text()


def exceptional_task(msg):
    raise RuntimeError(msg)


def test_task_exception(tmp_path, test_project):
    msg = 'all was going well except then the exceptional task excepted'
    task = minkipy.task(exceptional_task, args=(msg,))

    with pytest.raises(RuntimeError):
        task.run()

    assert msg in task.log_file.read_text()


def test_task_resubmit(tmp_path, test_project, test_queue):
    task = minkipy.task(exceptional_task, args=[
        'Arrrgg...',
    ])
    assert task.resubmit() is False

    test_queue.submit(task)

    with test_queue.next_task(timeout=2.) as incoming:
        incoming.run()

    assert task.queue == test_queue.name
    assert task.state == minkipy.FAILED
    assert 'Arrrgg...' in task.error

    task.resubmit()
    assert task.state == minkipy.QUEUED
    assert task in test_queue
    assert not task.error

    with test_queue.next_task(timeout=2.) as incoming:
        incoming.run()

    # Now try resubmitting on different queue (priority)
    priority = minkipy.queue('priority')
    task.resubmit(queue='priority')
    assert task.state == minkipy.QUEUED
    assert task.queue == 'priority'
    assert task in priority
    assert not task.error
    with minkipy.queue('priority').next_task(timeout=2.) as incoming:
        incoming.run()


def add_numbers(filename: Union[str, pathlib.Path]):
    with open(str(filename), 'r') as file:
        numbers = [int(line.rstrip()) for line in file.readlines()]
    return sum(numbers)


def test_task_files(tmp_path, test_project):
    # Have to do this as pytest returns pathlib2 Paths in py<=3.5.  This breaks other parts of the
    # code that expect a pathlib.Path. See:
    # https://github.com/pytest-dev/pytest/issues/5017
    tmp_path = pathlib.Path(str(tmp_path))

    TEST_FILE = tmp_path / 'numbers.dat'
    # Run the task in a subdirectory so it doesn't accidentally pick up the file
    TASK_PATH = tmp_path / 'task'

    with open(str(TEST_FILE), 'w') as file:
        file.write('\n'.join([str(num) for num in range(100)]))

    expected_result = add_numbers(TEST_FILE)

    task = minkipy.Task(minkipy.command(add_numbers, args=(TEST_FILE.name,)),
                        folder=str(TASK_PATH),
                        files=[TEST_FILE])
    assert len(task.files) == 1
    assert task.files[0].filename == TEST_FILE.name  # pylint: disable=unsubscriptable-object
    assert task.run() == expected_result

    # Now try adding the files manually
    task = minkipy.Task(
        minkipy.command(add_numbers, args=(TEST_FILE.name,)),
        folder=str(TASK_PATH),
    )
    task.add_files(TEST_FILE)
    assert len(task.files) == 1
    assert task.files[0].filename == TEST_FILE.name  # pylint: disable=unsubscriptable-object
    assert task.run() == expected_result


def test_task_parameters(test_project):
    """Make sure that the task() helper create the task correctly"""
    task = minkipy.task(my_task,
                        args=(1, 2, 3),
                        kwargs=dict(kword='this'),
                        dynamic=True,
                        folder='some_folder',
                        files=(__file__,))
    assert isinstance(task.cmd, minkipy.PythonCommand)
    assert task.cmd.dynamic is True
    assert tuple(task.cmd.args) == (1, 2, 3)
    assert task.cmd.kwargs == dict(kword='this')
    assert task.folder == 'some_folder'
    assert task.files[0].filename == os.path.basename(__file__)  # pylint: disable=unsubscriptable-object

    # Now make sure we can save the task with all parameters
    task_id = task.save()
    del task
    gc.collect()

    # Load and check
    task = mincepy.load(task_id)
    assert isinstance(task.cmd, minkipy.PythonCommand)
    assert task.cmd.dynamic is True
    assert tuple(task.cmd.args) == (1, 2, 3)
    assert task.cmd.kwargs == dict(kword='this')
    assert task.folder == 'some_folder'
    assert task.files[0].filename == os.path.basename(__file__)  # pylint: disable=unsubscriptable-object



import kiwipy

from . import queues

__all__ = ('run',)


def run(queue: queues.Queue, max_tasks: int = -1, timeout=60.) -> int:
    """
    Process a number of tasks from the given queue

    :param queue: the queue to process tasks from
    :param max_tasks: the maximum number of tasks to process
    :param timeout: the maximum time (in seconds) to wait for a new task
    """
    num_processed = 0
    try:
        while True:
            with queue.next_task(timeout=timeout) as fetched:
                fetched.run()
            num_processed += 1
            if max_tasks > 0 and num_processed >= max_tasks:
                return num_processed
    except kiwipy.QueueEmpty:
        return num_processed



# -*- coding: utf-8 -*-
import json
import os
import pathlib
from typing import Optional

import click
import kiwipy

__all__ = 'get_communicator', 'settings_path', 'ENV_MINKIPY_SETTINGS'

# pylint: disable=global-statement

ACTIVE_PROJECT_KEY = 'active_project'
ENV_MINKIPY_SETTINGS = 'MINKIPY_SETTINGS'
DEFAULT_SETTINGS = {}

# pylint: disable=invalid-name
_communicator = None


def get_communicator() -> Optional[kiwipy.Communicator]:
    global _communicator
    return _communicator


def set_communicator(communicator: kiwipy.Communicator):
    global _communicator
    _communicator = communicator


def read_settings() -> dict:
    """Read the current settings dictionary from disk"""
    path = settings_path()
    if not path.exists():
        write_settings(DEFAULT_SETTINGS)
        return DEFAULT_SETTINGS

    with open(str(path), 'r') as file:
        return json.load(file)


def write_settings(settings: dict):
    """Write a settings dictionary to the standard settings path"""
    path = settings_path()
    if not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)

    with open(str(settings_path()), 'w') as file:
        json.dump(settings, file, indent=4)


def settings_path() -> pathlib.Path:
    """Get the path to the settings file.  This will check the environment variable ENV_MINKIPY_SETTINGS for the
    settings and if this is not found will fall back to a sensible default."""
    try:
        return pathlib.Path(os.environ[ENV_MINKIPY_SETTINGS])
    except KeyError:
        app_dir = pathlib.Path(click.get_app_dir('minkipy', roaming=False))
        return app_dir / 'settings.json'



# -*- coding: utf-8 -*-
"""Tests of the settings module"""

import shutil

from minkipy import settings


def test_saving_settings(tmp_path):
    settings_dict = settings.read_settings()

    # Now, make sure that even if the folder is not present saving settings will still work
    # (by creating the folder first)
    shutil.rmtree(str(tmp_path))

    # This should succeed
    settings.write_settings(settings_dict)

    assert settings_dict == settings.read_settings()


def test_settings_path(monkeypatch):
    """Test that settings fall back to reasonable value even if the environmental variable
    isn't set"""
    monkeypatch.delenv(settings.ENV_MINKIPY_SETTINGS)
    assert settings.settings_path()



import minkipy


def add(a, b):
    return a + b


def test_create_task(tmp_path, test_project, queue_name):
    with minkipy.utils.working_directory(tmp_path):
        test_queue = minkipy.queue(queue_name)
        t1 = minkipy.task(add, (4, 5))
        t2 = minkipy.task(add, (24, 56))
        test_queue.submit(t1, t2)

        minkipy.run(test_queue, 2)

        assert t1.state == minkipy.DONE
        assert t2.state == minkipy.DONE


def test_empty(test_project, queue_name):
    test_queue = minkipy.queue(queue_name)
    assert minkipy.run(test_queue) == 0



from typing import Optional
import uuid

try:
    import pyos
except ImportError:
    pyos = None

import kiwipy
import mincepy

from . import settings

__all__ = ('Project', 'workon', 'project', 'working_on', 'get_active_project', 'set_active_project',
           'get_projects')

# pylint: disable=global-statement, redefined-outer-name

# pylint: disable=invalid-name
_working_on = None  # type: Optional[Project]
PROJECTS_KEY = 'projects'


def _make_default_queue_name(project_name) -> str:
    return "{}-default-queue".format(project_name)


class Project:

    def __init__(self, name: str):
        """Create a new project"""
        self.name = name
        self._uuid = uuid.uuid4()
        # Kiwipy settings
        self.kiwipy = {'connection_params': 'amqp://guest:guest@127.0.0.1/{}'.format(name)}
        # Mincepy settings
        self.mincepy = {'connection_params': 'mongodb://127.0.0.1/{}'.format(name)}
        self.default_queue = _make_default_queue_name(name)

    def __repr__(self) -> str:
        return "Project('{}')".format(self.name)

    @property
    def uuid(self) -> uuid.UUID:
        return self._uuid

    @classmethod
    def from_dict(cls, project_dict):
        # pylint: disable=protected-access
        project = Project.__new__(Project)

        project.name = project_dict['name']
        project.kiwipy = project_dict['kiwipy']
        project.mincepy = project_dict['mincepy']
        project._uuid = uuid.UUID(project_dict['uuid'])
        project.default_queue = project_dict.get('default_queue',
                                                 _make_default_queue_name(project.name))

        return project

    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'uuid': str(self.uuid),
            'kiwipy': self.kiwipy,
            'mincepy': self.mincepy,
            'default_queue': self.default_queue
        }

    def workon(self):
        historian = mincepy.create_historian(self.mincepy['connection_params'])
        kiwi_params = self.kiwipy['connection_params']
        if isinstance(kiwi_params, str):
            comm = kiwipy.connect(kiwi_params)
        elif isinstance(kiwi_params, dict):
            comm = kiwipy.connect(**kiwi_params)
        else:
            raise ValueError(
                "kiwi parameters must be string or dictionary, got '{}'".format(kiwi_params))
        settings.set_communicator(comm)

        mincepy.set_historian(historian)
        if pyos is not None:
            pyos.lib.init()

        _set_working_on(self)

    def save(self):
        """Save the project to the settings file"""
        settings_dict = settings.read_settings()
        settings_dict[PROJECTS_KEY][self.name] = self.to_dict()
        settings.write_settings(settings_dict)

    def set_as_active(self):
        settings_dict = settings.read_settings()
        # Make sure we're up to date
        settings_dict[PROJECTS_KEY][self.name] = self.to_dict()
        settings_dict[settings.ACTIVE_PROJECT_KEY] = self.name
        settings.write_settings(settings_dict)


def get_projects() -> dict:
    settings_dict = settings.read_settings()
    return {
        name: Project.from_dict(value) for name, value in settings_dict.get('projects', {}).items()
    }


def project(project_name: str = 'default') -> Project:
    """Fetch a project with a given name or create a new one"""
    settings_dict = settings.read_settings()
    update_settings_file = False

    stored = settings_dict.setdefault('projects', {})

    if project_name not in stored:
        proj = Project(project_name)
        stored[project_name] = proj.to_dict()
        settings_dict.setdefault(settings.ACTIVE_PROJECT_KEY, project_name)
        update_settings_file = True
    else:
        proj = Project.from_dict(stored[project_name])

    if update_settings_file:
        settings.write_settings(settings_dict)

    return proj


def workon(project_name: str = None, auto_create=False) -> Optional[Project]:
    """
    Set the project currently being worked on.  This will set the global historian and kiwipy
    communicator so be careful if you were already using these.

    :param project_name: the project to work on.  If None, will use the currently active project
    :param auto_create: if True automatically create a project if it doesn't exist
    """
    global _working_on
    if project_name is None:
        settings_dict = settings.read_settings()
        project_name = settings_dict[settings.ACTIVE_PROJECT_KEY]

    proj = get_projects().get(project_name, None)
    if proj is None:
        if auto_create:
            proj = project(project_name)
        else:
            raise ValueError("Project '{}' does not exist.".format(project_name))

    if _working_on is not None and _working_on.uuid == proj.uuid:
        return proj

    proj.workon()
    return proj


def working_on() -> Project:
    global _working_on
    if _working_on is None:
        workon()
    return _working_on


def get_active_project() -> Optional[Project]:
    """Gets the currently active project from the settings.  This may be not be the project
    currently being worked on"""
    settings_dict = settings.read_settings()
    active = settings_dict.get(settings.ACTIVE_PROJECT_KEY, None)
    if active is None:
        return None

    return Project.from_dict(settings_dict['projects'][active])


def set_active_project(name):
    """Set the currently active project in the settings"""
    settings_dict = settings.read_settings()
    if name not in settings_dict['projects']:
        raise ValueError("Project '{}' does not exist".format(name))
    settings_dict[settings.ACTIVE_PROJECT_KEY] = name
    settings.write_settings(settings_dict)


def _set_working_on(project: Project):
    global _working_on
    _working_on = project


def _get_working_on() -> Project:
    global _working_on
    return _working_on



# -*- coding: utf-8 -*-
import collections
import contextlib
import functools
import logging
from typing import Iterator, Any, Sequence, Dict, Union

import beautifultable
import kiwipy.rmq
from kiwipy.rmq import threadcomms
import mincepy

try:
    import pyos
    import pyos.pathlib
except ImportError:
    pyos = None

import minkipy  # pylint: disable=unused-import
from . import projects
from . import settings
from . import tasks

__all__ = 'Queue', 'queue'

logger = logging.getLogger(__name__)

TASK_ID = 'task_id'


class Queue:
    """A task queue"""

    def __init__(self,
                 communicator: kiwipy.rmq.RmqThreadCommunicator,
                 historian: mincepy.Historian,
                 queue_name='default-queue'):
        self._communicator = communicator
        self._historian = historian
        self._kiwi_queue = communicator.task_queue(queue_name)
        self._name = queue_name

    def size(self):
        count = 0
        for _ in self._kiwi_queue:
            count += 1
        return count

    def __iter__(self) -> Iterator[tasks.Task]:
        """Iterate through the tasks in this queue in the order they were submitted"""
        for msg in self._kiwi_queue:
            yield self._historian.load(msg.body[TASK_ID])

    def __contains__(self, item: Union[tasks.Task, Any]) -> bool:
        obj_id = self._historian.to_obj_id(item)

        for msg in self._kiwi_queue:
            if obj_id == msg.body[TASK_ID]:
                return True

        return False

    def __str__(self) -> str:
        """Get the name of this queue"""
        return self.name

    @property
    def name(self) -> str:
        return self._name

    def empty(self) -> bool:
        """Returns True if the queue is empty, False otherwise"""
        try:
            with self._kiwi_queue.next_task(timeout=0.01, fail=True):
                return False
        except kiwipy.QueueEmpty:
            return True

    def list(self, verbosity: int = 1):
        """Pretty-print the list of tasks."""

        def load_generator():
            for incoming in self._kiwi_queue:
                yield self._historian.load(incoming.body[TASK_ID])

        pprint(load_generator(), verbosity)

    @contextlib.contextmanager
    def next_task(self, timeout=None):
        """Get the next task from the queue.

        :param timeout: the duration (in seconds) to wait for a task to become available
        """
        with self._kiwi_queue.next_task(timeout=timeout) as ktask:
            with ktask.processing() as outcome:
                msg = ktask.body
                task = self._historian.load(msg[TASK_ID])  # type: tasks.Task
                task.state = tasks.PROCESSING

                try:
                    yield task
                    # Task done
                    if task.state == tasks.RUNNING:
                        task.state = tasks.DONE
                except Exception as exc:  # pylint: disable=broad-except
                    outcome.set_exception(exc)
                else:
                    outcome.set_result(True)

    def submit(self, *tasks: 'minkipy.Task'):  # pylint: disable=redefined-outer-name
        """Submit one or more tasks to the queue.  The task ids will be returned."""
        task_ids = []

        # First check if any of the passed tasks are already in queues in which case we have to
        # remove them

        existing_queues = collections.defaultdict(list)
        for entry in tasks:
            # Don't try to remove ones in this queue, this will be checked for on submitting
            if entry.queue and entry.queue != self._name:
                existing_queues[entry.queue].append(entry)

        for name, queued_task in existing_queues.items():
            queue(name).remove(*queued_task)

        # Get the ids of the current tasks
        current_ids = {msg.body[TASK_ID] for msg in self._kiwi_queue}
        already_queued = []
        for task in tasks:
            task.save()
            if task.obj_id in current_ids:
                already_queued.append(task.obj_id)
            else:
                task_ids.append(self.submit_one(task))

        if already_queued:
            logger.warning('Skipping the following tasks because they are already in the queue: %s',
                           already_queued)

        if len(tasks) == 1:
            if not task_ids:
                return None

            return task_ids[0]

        return task_ids

    def submit_one(self, task: tasks.Task) -> Any:
        """Submit one task to the queue.  The object id for the task will be returned."""
        # Encode the task
        task.save()
        task_id = task.obj_id
        msg = {TASK_ID: task_id}
        self._kiwi_queue.task_send(msg, no_reply=True)
        task.queue = self.name
        task.state = tasks.QUEUED
        return task_id

    def remove(self, *task: Union[tasks.Task, Any]) -> list:
        """Remove a task from the queue.  Can supply the task instance or the object id of the task.
        Returns a list of the object ids of the removed tasks
        """
        obj_ids = set(map(self._historian.to_obj_id, task))
        removed = []
        for kiwi_task in self._kiwi_queue:
            oid = kiwi_task.body[TASK_ID]
            if oid in obj_ids:
                self._drop_task(kiwi_task)
                obj_ids.remove(oid)
                removed.append(oid)

            if not obj_ids:
                break

        return removed

    def purge(self) -> int:
        """Cancel all tasks in this queue"""
        num_cancelled = 0
        for kiwi_task in self._kiwi_queue:
            self._drop_task(kiwi_task)
            num_cancelled += 1

        return num_cancelled

    def _drop_task(self, kiwi_task: threadcomms.RmqThreadIncomingTask):
        task = self._historian.load(kiwi_task.body[TASK_ID])  # type: tasks.Task

        # There is a bug in kiwipy that prevents us from cancelling this future so
        # for now just set a cancelled results.  In any case the result is not sent
        # back for the time being.
        with kiwi_task.processing() as outcome:
            outcome.set_result('Cancelled')

        task.queue = ''
        task.state = tasks.CANCELED
        task.save()
        return task


def queue(name: str = None,
          communicator: kiwipy.Communicator = None,
          historian: mincepy.Historian = None):
    """Get a queue of the given name.  If the queue doesn't exist it will be
    created.  If None is passed the default queue will be used.
    """
    if name is None:
        name = projects.working_on().default_queue

    communicator = communicator or settings.get_communicator()
    historian = historian or mincepy.get_historian()
    return Queue(communicator, historian, name)


def pprint(tasks_list: Iterator[tasks.Task], verbosity: int = 2) -> None:
    """Pretty print information about tasks"""
    if verbosity < 0:
        return

    if not tasks_list:
        print('Empty')
        return

    headers = ['obj_id', 'cmd', 'state', 'error']
    col_widths = [26, 38, max(map(len, tasks.STATES)), 16]
    col_align = [
        beautifultable.ALIGN_RIGHT, beautifultable.ALIGN_LEFT, beautifultable.ALIGN_RIGHT,
        beautifultable.ALIGN_LEFT
    ]

    def fetch(name, obj, transform=str):
        return transform(getattr(obj, name))

    # Create the getter functions that will fetch the data
    getters = [functools.partial(fetch, col) for col in headers]
    if pyos is not None:
        headers.insert(1, 'pyos_path')
        getters.insert(1, functools.partial(fetch, 'pyos_path'))
        col_widths.insert(1, 26)
        col_align.insert(1, beautifultable.ALIGN_LEFT)
        pyos_paths = collections.defaultdict(int)  # type: Dict[str, int]

    table = _create_table()
    table.columns.header = headers
    table.columns.width = col_widths
    table.columns.alignment = col_align

    state_counts = collections.defaultdict(int)  # type: Dict[str, int]

    def get_row(task) -> Sequence:
        row = []
        for getter in getters:
            value = getter(task)
            row.append(value)
        state_counts[task.state] += 1
        if pyos is not None:
            pyos_paths[task.pyos_path] += 1
        state_counts['total'] += 1
        return row

    for line in table.stream(map(get_row, tasks_list)):
        if verbosity >= 1:
            print(line)

    if state_counts['total'] == 0:
        print('Empty')
    else:
        if pyos is not None:
            for path, count in pyos_paths.items():
                print('Tasks in {}: {}'.format(path, count))
        print(', '.join('{}: {}'.format(state, count) for state, count in state_counts.items()))


def _create_table() -> beautifultable.BeautifulTable:
    """Creates a new table for printing"""
    table = beautifultable.BeautifulTable()
    table.set_style(beautifultable.STYLE_COMPACT)
    table.columns.width_exceed_policy = beautifultable.WEP_ELLIPSIS

    return table



# -*- coding: utf-8 -*-
from contextlib import contextmanager, redirect_stdout, redirect_stderr
try:
    from contextlib import nullcontext
except ImportError:
    from contextlib2 import nullcontext
import logging
import os
import uuid
import pathlib
import sys
from typing import List, Sequence

import mincepy

try:
    import pyos
except ImportError:
    pyos = None

import minkipy
from . import commands
from . import utils

__all__ = ('CREATED', 'QUEUED', 'HELD', 'RUNNING', 'DONE', 'FAILED', 'CANCELED', 'TIMEOUT',
           'MEMORY', 'Task', 'task')

# Possible states
CREATED = 'created'
QUEUED = 'queued'
HELD = 'held'
PROCESSING = 'processing'
RUNNING = 'running'
DONE = 'done'
FAILED = 'FAILED'
CANCELED = 'CANCELED'
TIMEOUT = 'TIMEOUT'
MEMORY = 'MEMORY'

STATES = [CREATED, QUEUED, HELD, PROCESSING, RUNNING, DONE, FAILED, CREATED, TIMEOUT, MEMORY]

logger = logging.getLogger(__name__)


class Task(mincepy.SimpleSavable):
    """A minkiPy task.  This represents a unit of work that can be submitted to a queue."""

    # pylint: disable=too-many-instance-attributes
    TYPE_ID = uuid.UUID('bc48616e-4fcb-41b2-bd03-a37a8fe1dce7')

    folder = mincepy.field()
    error = mincepy.field()
    queue = mincepy.field()
    log_level = mincepy.field()

    def __init__(self,
                 cmd: commands.Command,
                 folder: str,
                 files: List[pathlib.Path] = (),
                 historian: mincepy.Historian = None):
        """Create a new task

        :param cmd: the command to be executed
        :param folder: the path to the folder where the task will be run
        :param files: a list of dependent files to include with the task
        :param historian: an optional historian to use (will use mincepy.get_historian() if not
            supplied)
        """
        assert isinstance(folder,
                          str), "Folder name must be a string, got '{}'".format(type(folder))
        super().__init__()
        self._historian = historian or mincepy.get_historian()

        self._cmd = cmd
        self.folder = folder  # The name of the folder where the task will be run
        self._files = mincepy.builtins.RefList()
        for file in files:
            self.add_files(file)
        self._state = CREATED
        self.error = ''
        self.queue = ''  # Set the the name of the queue it's in if it gets put in one
        self.log_level = logging.WARNING
        self._log_file = self._historian.create_file('task_log', encoding='utf-8')
        self._stdout = self._historian.create_file('stdout', encoding='utf-8')
        self._stderr = self._historian.create_file('stderr', encoding='utf-8')

        if pyos is not None:
            self._pyos_path = pyos.os.getcwd()  # type: str
        else:
            self._pyos_path = None

    def __str__(self) -> str:
        str_list = []
        str_list.append('state={}'.format(self._state))
        if self.error:
            str_list.append('[{}]'.format(self.error))
        return ' '.join(str_list)

    @mincepy.field('_state')
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value
        self.save()

    @mincepy.field('_cmd')
    def cmd(self) -> minkipy.Command:
        return self._cmd

    @mincepy.field('_files')
    def files(self) -> Sequence[mincepy.builtins.File]:
        return self._files

    @mincepy.field('_log_file')
    def log_file(self) -> mincepy.builtins.File:
        """Get the log file"""
        return self._log_file

    @mincepy.field('_stdout')
    def stdout(self) -> mincepy.builtins.File:
        """Get the standard out file"""
        return self._stdout

    @mincepy.field('_stderr')
    def stderr(self) -> mincepy.builtins.File:
        """Get the standard err file"""
        return self._stderr

    @mincepy.field('_pyos_path')
    def pyos_path(self):
        if self._pyos_path is None:
            return None

        if pyos is None:
            # Fallback if we have no pyos: just return the string
            return self._pyos_path

        return pyos.pathlib.PurePath(self._pyos_path)

    @pyos_path.setter
    def pyos_path(self, new_path):
        if new_path is None:
            self._pyos_path = None
        else:
            self._pyos_path = str(new_path)

    def add_files(self, filename: [str, pathlib.Path]):
        filename = pathlib.Path(filename)
        file = self._historian.create_file(filename.name)
        file.from_disk(filename)
        self._files.append(file)

    def run(self):
        with self._capture_log(), self._capture_stds():
            logger.info('Starting task with id %s', self.obj_id)
            if pyos and self.pyos_path is not None:
                path_context = pyos.pathlib.working_path(self.pyos_path)
                logger.debug("Running in pyos path '%s'", self.pyos_path)
            else:
                logger.debug('Running without pyos')
                path_context = nullcontext()

            with path_context:
                try:
                    self.state = RUNNING
                    if self.folder and not os.path.exists(self.folder):
                        os.makedirs(self.folder)
                    self.copy_files_to(self.folder)

                    # Change the directory to the running folder and back at the end
                    with utils.working_directory(self.folder):
                        result = self._cmd.run()

                    self._state = DONE
                    return result
                except Exception as exc:
                    logger.exception("Task '%s' excepted", self.obj_id)
                    self.error = str(exc)
                    self.state = FAILED
                    raise
                finally:
                    self.save()

    def resubmit(self, queue='') -> bool:
        """Resubmit this task if it has already been submitted before.

        Will return `True` if the task was successfully resubmitted.
        If the task has never been submitted or is still in the queue `False` will be returned.

        :param queue: a queue name that can be supplied to resubmit on a particular queue, if not
            supplied original queue that this task was submitted on will be used.
        """
        if not self.queue:
            # Has not been submitted to a queue
            return False

        with self._capture_log():
            self.error = ''  # Reset this
            try:
                submit_to = queue or self.queue
                result = minkipy.queue(submit_to).submit(self)
            except Exception:
                self.sync()  # Reset our state
                raise
            else:
                if result is None:
                    logger.info("Failed to resubmit task '%s'.", self.obj_id)
                    self.sync()  # Reset state back to what it was
                else:
                    logger.info("Resubmitted task '%s'.", self.obj_id)

        return result is not None

    def copy_files_to(self, folder):
        """Copy the task files to the given folder.  The folder must exist already."""
        # For ease of debugging write any command file (e.g. the script file) as well
        # do this first just in case any of the files have a file with the same name
        self._cmd.copy_files_to(folder)
        for file in self._files:
            file.to_disk(folder)

    @contextmanager
    def _capture_log(self):
        """Context handler to enable loggin' on the task"""
        if self.log_level is None:
            # Don't save the log
            yield
            return

        root_logger = logging.getLogger()  # Get the top level logger
        with self._log_file.open(mode='a') as file:
            root_logger.setLevel(self.log_level)
            handler = logging.StreamHandler(file)

            handler.setLevel(self.log_level)
            handler.setFormatter(
                logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

            # Check if the root logger is currently set to a higher level than the level
            # we want to log at which would mean we don't get the log messages.  If so,
            # temporarily lower it.  See here for an explanation of why this is necessary:
            # https://www.electricmonk.nl/log/2017/08/06/understanding-pythons-logging-module/
            original_root_level = None
            if root_logger.level > self.log_level:
                original_root_level = root_logger
                root_logger.setLevel(self.log_level)

            try:
                root_logger.addHandler(handler)
                yield
            finally:
                root_logger.removeHandler(handler)
                if original_root_level is not None:
                    root_logger.setLevel(original_root_level)

    @contextmanager
    def _capture_stds(self):
        """Capture standard out and err"""
        with self._stdout.open(mode='a') as stdout, self._stderr.open(mode='a') as stderr:
            out = utils.TextMultiplexer(sys.stdout, stdout)
            err = utils.TextMultiplexer(sys.stderr, stderr)
            with redirect_stdout(out), redirect_stderr(err):
                yield


def task(
        cmd,
        args=(),
        kwargs=None,
        dynamic=False,
        folder: str = '',
        files=(),
):
    """Create a task

    :param cmd: the command for the task to execute
    :param args: the argument to the command
    :param kwargs: the keyword arguments to the command (not all command types support this)
    :param folder: the path where the task should run (can be absolute or relative)
    :param files: an optional list of files for the task to copy
    :param dynamic: a flag to make the command dynamic, not all commands types support this
    """
    return Task(commands.command(cmd, args, kwargs=kwargs, dynamic=dynamic), folder, files=files)


HISTORIAN_TYPES = (Task,)



# -*- coding: utf-8 -*-
import minkipy

# pylint: disable=unused-argument


def do_stuff(arg):
    return arg


def test_queue_basic(tmp_path, test_project, queue_name):
    with minkipy.utils.working_directory(tmp_path):
        queue = minkipy.queue(queue_name)
        assert str(queue) == queue_name
        assert queue.empty()  # Should have nothing in it yet

        task = minkipy.task(do_stuff, ['stuff'])
        queue.submit(task)

        assert not queue.empty()

        with queue.next_task(timeout=2.) as fetched:
            assert fetched.run() == 'stuff'

        # Now check our original 'task' is finished
        task.sync()
        assert task.state == 'done', task.error


def test_queue_iter(test_project, queue_name):
    queue = minkipy.queue(queue_name)

    for idx in range(10):
        task = minkipy.task(do_stuff, [idx])
        queue.submit(task)

    for idx, queued in enumerate(queue):
        assert queued.cmd.args[0] == idx

    assert queue.size() == 10
    queue.purge()
    assert queue.size() == 0


def test_double_submission(test_project, test_queue: minkipy.Queue):
    """Test that you can't submit the same task twice"""
    task = minkipy.task(do_stuff, [None])
    task.save()
    submitted = test_queue.submit(task)
    assert submitted == task.obj_id

    # Should be skipped
    assert not test_queue.submit(task)


def test_remove(test_project, test_queue: minkipy.Queue):
    task = minkipy.task(do_stuff, [None])
    obj_id = test_queue.submit(task)
    assert task in test_queue

    removed = test_queue.remove(task)
    assert removed == [obj_id]
    assert task not in test_queue


def test_submit_to_different_queue(test_project, test_queue: minkipy.Queue):
    task = minkipy.task(do_stuff, [None])
    test_queue.submit(task)
    assert task in test_queue

    second_queue = minkipy.queue('second-test-queue')
    second_queue.submit(task)
    assert task not in test_queue
    assert task in second_queue


def test_queue_contains(test_project, test_queue: minkipy.Queue):
    task = minkipy.task(do_stuff, [None])
    task_id = test_queue.submit(task)
    # Check the different possibilities for testing if a task is in a queue
    assert task in test_queue
    assert task_id in test_queue
    assert str(task_id) in test_queue



"""This module acts as a shim to call a python command dynamically i.e. import the file when a
task is ran as opposed to having the function statically stored in the task.

This is useful when you want new tasks to pick up any code changes you make after you submit them

DEV WARNING: Do no change the location, signature or name of run_dynamically as it may break things
for users that have saved tasks (with the current signature)
"""

from minkipy import utils


def run_dynamically(_fn_spec: tuple, *args, **kwargs):
    """Run the given function from the script file passing the args and kwargs and returning the
    result"""
    script_file, function = _fn_spec
    script = utils.load_script(script_file)
    print("Got script {}".format(script))
    run = utils.get_symbol(script, function)
    print("Got run: {}".format(run))
    kwargs = kwargs or {}
    return run(*args, **kwargs)



import pprint

import click

import minkipy
from . import main

pretty = pprint.PrettyPrinter(depth=4)  # pylint: disable=invalid-name


@main.minki.group('project')
def project_():
    pass


@project_.command()
@click.argument('name', type=str)
def create(name):
    project = minkipy.project(name)
    click.echo("Created project '{}'".format(name))
    click.echo(pretty.pformat(project.to_dict()))


@project_.command()
def list():  # pylint: disable=redefined-builtin
    active = minkipy.get_active_project()
    for project in minkipy.get_projects().values():
        line = [project.name]
        if project.uuid == active.uuid:
            line.append('[active]')

        click.echo(" ".join(line))


@project_.command()
@click.argument('project', type=str, default=None, required=False)
def show(project):
    """Show the project settings"""
    if project is None:
        proj = minkipy.get_active_project()
    else:
        proj = minkipy.get_projects().get(project, None)

    click.echo("{}:".format(proj.name))

    if proj is None:
        click.echo("Project not found")
        return 1

    click.echo(pretty.pformat(proj.to_dict()))
    return 0



# pylint: disable=invalid-name

author_info = (('Martin Uhrin', 'martin.uhrin.10@ucl.ac.uk'),)
version_info = (0, 2, 21)

__author__ = ", ".join("{} <{}>".format(*info) for info in author_info)
__version__ = ".".join(map(str, version_info))

__all__ = ('__version__',)



# Test pyos related functionality

import pytest

# Skip these tests if pyos is not installed
pyos = pytest.importorskip("pyos")  # pylint: disable=invalid-name
import pyos.pyos
from mincepy.testing import Car

import minkipy


def car_maker():
    return Car().save()


def test_pyos_path_default(tmp_path, test_project):
    # Create a task from a function in a script
    cwd = pyos.pyos.pwd()
    with minkipy.utils.working_directory(tmp_path):
        task = minkipy.task(car_maker)
        assert task.pyos_path == cwd
        car_id = task.run()
        # Now look for the car
        results = pyos.pyos.ls(car_id)
        assert len(results) == 1
        assert results[0].obj_id == car_id
        assert isinstance(pyos.pyos.load(results[0]), Car)


def test_pyos_path_custom(tmp_path, test_project):
    # Create a task from a function in a script
    task_path = pyos.Path('work/here/')
    with minkipy.utils.working_directory(tmp_path):
        task = minkipy.task(car_maker)
        task.pyos_path = task_path

        car_id = task.run()
        # Now look for the car
        results = pyos.pyos.ls(task_path / str(car_id))
        assert len(results) == 1
        assert results[0].obj_id == car_id
        assert isinstance(pyos.pyos.load(results[0]), Car)



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
import inspect
from pathlib import Path
import sys
from typing import List, Optional, Sequence, Mapping, Union
import uuid

import mincepy

from . import constants
from . import utils

__all__ = 'Command', 'command', 'PythonCommand'

MODULE_PREFIX = ':mod:'


class Command(mincepy.SimpleSavable, metaclass=ABCMeta):
    TYPE_ID = uuid.UUID('38dc3093-058f-4934-9a48-292eeef35e11')

    def __init__(self, args: Sequence):
        super().__init__()
        self._args = mincepy.RefList(args)

    @mincepy.field('_args')
    def args(self) -> tuple:
        return self._args

    @abstractmethod
    def run(self) -> Optional[List]:
        """Run the command with the stored arguments"""

    def copy_files_to(self, path):
        """Copy any command files to the given path"""


def command(
        cmd,
        args: Sequence = (),
        type: str = 'python-function',  # pylint: disable=redefined-builtin
        **kwargs) -> Command:
    """Command creation factory.

    :param cmd: the command specification, for python function this should be a function or class
        method
    :param args: the (positional) argument for the command
    :param type: the command type, default is 'python-function'
    :param kwargs: additional arguments that will be passed as kwargs to the relevant command class
        constructor so have a look at it's constructor for more details
    """
    if type == 'python-function':
        return PythonCommand.build(cmd, args, **kwargs)

    raise ValueError("Unknown command type '{}'".format(type))


class PythonCommand(Command):
    TYPE_ID = uuid.UUID('61736206-729b-4a0b-9fac-6b5e71123ba0')

    @classmethod
    def build(cls, cmd, args: Sequence = (), dynamic=False, kwargs: dict = None, **rest):
        function = 'run'  # The default function name

        if inspect.ismethod(cmd):
            if dynamic:
                script_file = constants.MODULE_PREFIX + cmd.__module__
            else:
                script_file = sys.modules[cmd.__module__].__file__
            function = utils.get_symbol_name(cmd)
            args = (cmd.__self__,) + args
        elif inspect.isfunction(cmd):
            if dynamic:
                script_file = constants.MODULE_PREFIX + cmd.__module__
            else:
                script_file = sys.modules[cmd.__module__].__file__
            function = cmd.__name__
        elif inspect.ismodule(cmd):
            if dynamic:
                script_file = constants.MODULE_PREFIX + cmd.__name__
            else:
                script_file = cmd.__file__
        elif isinstance(cmd, str) and '@' in cmd:
            script_file, function = cmd.split('@')
            if not script_file.endswith('.py'):
                # Assume they've given a module path
                script_file = sys.modules[script_file].__file__
        else:
            raise ValueError("Unknown python function command '{}".format(cmd))

        return PythonCommand(script_file, function, args, dynamic=dynamic, kwargs=kwargs, **rest)

    # pylint: disable=too-many-arguments
    def __init__(self,
                 script_file,
                 function='run',
                 args=(),
                 kwargs: dict = None,
                 dynamic=False,
                 historian=None):
        """
        Create a python command

        :param script_file: the path to the script file
        :param function: the name of the function in the script file to invoke
        :param args: the arguments to the function
        :param kwargs: the keyword arguments to the function
        :param dynamic: whether to run the function dynamically
            (i.e. import it when the command is ran)
        :param historian: the historian
        """
        super().__init__(args)
        self._historian = historian or mincepy.get_historian()

        if dynamic:
            self._script_file = script_file
        else:
            # Save the script
            script_file = Path(script_file)
            self._script_file = self._historian.create_file(script_file.name, 'utf-8')
            self._script_file.from_disk(script_file)

        self._dynamic = dynamic
        self._function = function
        self._kwargs = mincepy.RefDict(kwargs or {})

    def __str__(self):
        return '{}@{}{}'.format(self._script_file, self._function, self._args)

    @mincepy.field('_dynamic')
    def dynamic(self) -> bool:
        """If True then the task script file will be loaded dynamically and task.script_file will be a string,
        otherwise the file is stored directly in the task."""
        return self._dynamic

    @mincepy.field('_script_file')
    def script_file(self) -> Union[mincepy.File, str]:
        """Access the python script file.
        This can be either a :class:`mincepy.File` if it is stored directly in the task or a string.
        If it is a string it will either be the path to the file or a module path specified as
        ':mod:path.to.module' where 'module' would be imported"""
        return self._script_file

    @mincepy.field('_function')
    def fn_name(self) -> str:
        """The name of the function that will be run in the script"""
        return self._function

    @mincepy.field('_kwargs')
    def kwargs(self) -> Mapping:
        return self._kwargs

    def load_instance_state(self, saved_state, loader: 'mincepy.Loader'):
        super().load_instance_state(saved_state, loader)
        # Deal with new attributes that were added (in case we load an old record)
        if not hasattr(self, '_kwargs'):
            self._kwargs = {}
        if not hasattr(self, '_dynamic'):
            self._dynamic = False

    def run(self) -> Optional[List]:
        """Run this python command"""
        script = utils.load_script(self._script_file)
        run = utils.get_symbol(script, self._function)
        kwargs = self._kwargs or {}
        return run(*self._args, **kwargs)

    def copy_files_to(self, path):
        # Only copy the task file if it is static
        if not self.dynamic:
            self._script_file.to_disk(path)


HISTORIAN_TYPES = Command, PythonCommand



try:
    import cmd2

except ImportError:

    def get_commands():
        return []
else:
    import argparse
    import sys

    from . import projects

    class MinkiCommands(cmd2.CommandSet):
        workon_parser = argparse.ArgumentParser()
        workon_parser.add_argument('project', type=str)

        @cmd2.with_argparser(workon_parser)
        def do_workon(self, args):  # pylint: disable=no-self-use
            try:
                project = projects.workon(args.project)
            except ValueError as exc:
                print(exc, file=sys.stderr)
                return 1
            else:
                print("Switched to '{}'".format(project))

            return 0

    def get_commands():
        """Get all the pyos shell commands served by minkipy"""
        return [MinkiCommands()]



# -*- coding: utf-8 -*-
import importlib
import inspect
import os.path
import pathlib

import mincepy

import minkipy

# pylint: disable=invalid-name


def add(a, b, prefactor=1.):
    return prefactor * (a + b)


def test_basics():
    cmd = minkipy.PythonCommand.build(add, args=(5, 10), kwargs=dict(prefactor=2.))
    assert cmd.fn_name == 'add'  # pylint: disable=comparison-with-callable
    assert cmd.script_file.filename == os.path.basename(__file__)  # pylint: disable=no-member
    assert cmd.run() == 30
    assert cmd.run() == 30, 'Just making sure we can run multiple times'


def mul(a, b, prefactor=1.):
    return prefactor * a * b


def test_dynamic():
    """Test the dynamic execution of python commands"""
    MODULE_PATH = str(pathlib.Path(__file__).parent / 'test_module.py')
    source = inspect.getsource(mul)

    try:
        with open(MODULE_PATH, 'w') as file:
            file.write(source)

        # pylint: disable=import-outside-toplevel, no-name-in-module
        from . import test_module

        static = minkipy.PythonCommand.build(test_module.mul,
                                             args=(5, 10),
                                             kwargs=dict(prefactor=2.),
                                             dynamic=False)
        dynamic = minkipy.PythonCommand.build(test_module.mul,
                                              args=(5, 10),
                                              kwargs=dict(prefactor=2.),
                                              dynamic=True)

        source = inspect.getsource(add).replace('add', 'mul')
        with open(MODULE_PATH, 'w') as file:
            file.write(source)
        importlib.reload(test_module)  # Reload the module to register the change

        assert static.run() == 100.
        assert dynamic.run() == 30.
    finally:
        os.remove(MODULE_PATH)


def dummy(kwarg=None):
    return kwarg


def test_savable_kwargs():
    """Test putting mincepy savable objects in the keyword arguments"""
    car = mincepy.testing.Car(colour='red', make='ferrari')
    car_id = car.save()
    cmd = minkipy.PythonCommand.build(dummy, kwargs=dict(kwarg=car))
    cmd_id = cmd.save()

    # Delete everything, reload and check that we've still got the same car
    del car, cmd

    cmd = mincepy.load(cmd_id)
    assert cmd.kwargs['kwarg'].obj_id == car_id

    result = cmd.run()
    assert result.obj_id == car_id



# -*- coding: utf-8 -*-
import contextlib
import functools
import importlib.machinery
import importlib.util
import inspect
import io
import os
from pathlib import Path
import shutil
import tempfile
import types
from typing import Optional, Sequence

import mincepy

from . import constants

__all__ = ('load_script',)


@functools.singledispatch
def load_script(script_file: [str, io.TextIOBase]) -> types.ModuleType:
    """Load a module from the given script file.

    The script file can be an io.TextIOBase in which case it will be saved and loaded directly.
    If it is a mincepy.File it will be opened and loaded.
    If it is a string it will be interpreted as follows:
        * If is starts with :mod:... it will be treated as a module
        * If it starts with :file:... the file will be loaded as a module
        * Otherwise the string is treated as a file path as if it were prefixed with :file:
    """
    raise TypeError('Unknown script file type: {}'.format(script_file.__class__.__name__))


@load_script.register(str)
def _(script_file: str) -> types.ModuleType:
    if script_file.startswith(constants.MODULE_PREFIX):
        mod_name = script_file[len(constants.MODULE_PREFIX):]
        return importlib.import_module(mod_name)

    if script_file.startswith(constants.FILE_PREFIX):
        script_file = script_file[len(constants.FILE_PREFIX):]

    # Ok, assume that the string _is_ the script itself
    spec = importlib.util.spec_from_file_location('script', script_file)
    script = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(script)
    return script


@load_script.register(io.TextIOBase)
def _(script_file: io.TextIOBase) -> types.ModuleType:
    temp_file = None
    script_path = None
    try:
        # Save to a temporary file
        with tempfile.NamedTemporaryFile('w', encoding='utf-8', suffix='.py',
                                         delete=False) as temp_file:
            shutil.copyfileobj(script_file, temp_file)
            script_path = temp_file.name

        return load_script(constants.FILE_PREFIX + script_path)
    finally:
        if temp_file is not None:
            os.remove(script_path)


@load_script.register(mincepy.File)
def _(script_file: mincepy.File) -> types.ModuleType:
    with script_file.open('r') as file:
        return load_script(file)


def get_symbol(module, name: str):
    """Given a module and a string that represents a path to a symbol this will retrieve it
    e.g. name='MyClass.do_stuff' will get the do_stuff function from the class MyClass within
    the module 'module'"""
    path = name.split('.')
    obj = module
    for entry in path:
        obj = getattr(obj, entry)
    return obj


def yield_symbol_names(symbol):
    if inspect.ismethod(symbol) or inspect.isfunction(symbol):
        if inspect.ismethod(symbol):
            yield from yield_symbol_names(symbol.__self__)
        yield symbol.__name__
    elif isinstance(symbol, object):
        yield symbol.__class__.__name__
    else:
        raise TypeError("Unknown symbol type: '{}'".format(symbol))


def get_symbol_name(symbol):
    return '.'.join(yield_symbol_names(symbol))


@contextlib.contextmanager
def working_directory(path):
    """Changes working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    if path:
        os.chdir(str(path))  # str call for python < 3.6
    try:
        yield
    finally:
        os.chdir(str(prev_cwd))


class TextMultiplexer(io.TextIOBase):
    """Takes a primary output stream and sends all write command to it and all the secondary
    streams.  Otherwise behaves like the primary stream.
    """

    def __init__(self, primary, *secondary):
        super().__init__()
        self._primary = primary  # type: io.TextIOBase
        self._secondary = secondary  # type: Sequence[io.TextIOBase]

    # Deliberately don't implement close!

    @property
    def closed(self):
        return self._primary.closed

    @property
    def encoding(self):
        return self._primary.encoding

    @property
    def errors(self):
        return self._primary.errors

    def fileno(self) -> int:
        return self._primary.fileno()

    def flush(self) -> None:
        return self._primary.flush()

    def isatty(self) -> bool:
        return self._primary.isatty()

    @property
    def mode(self):
        return self._primary.mode

    @property
    def name(self):
        return self._primary.name

    def writable(self) -> bool:
        return self._primary.writable()

    @property
    def newlines(self):
        return self._primary.newlines

    def next(self):
        return self._primary.next()

    def read(self, size: Optional[int] = ...) -> str:
        return self._primary.read(size)

    def write(self, s: str):
        self._primary.write(s)
        for stream in self._secondary:
            stream.write(s)


HISTORIAN_TYPES = tuple()



# -*- coding: utf-8 -*-
from .commands import *
from .projects import *
from .queues import *
from .settings import *
from .tasks import *
from .utils import *
from .version import *
from .workers import *
# Keep the linter happy
from . import constants
from . import defaults
from . import projects
from . import pyos_extensions
from . import version
from . import workers

_ADDITIONAL = 'defaults', 'constants'

__all__ = (commands.__all__ + queues.__all__ + tasks.__all__ + utils.__all__ + version.__all__ +
           workers.__all__ + settings.__all__ + projects.__all__) + _ADDITIONAL



"""Extensions to mincePy and the mincePy GUI"""
from typing import Optional, Iterable

import mincepy

from . import projects
from . import queues
from . import tasks
from . import commands
from . import utils


def get_types():
    """Provide a list of all historian types"""
    types = list()
    types.extend(tasks.HISTORIAN_TYPES)
    types.extend(commands.HISTORIAN_TYPES)
    types.extend(utils.HISTORIAN_TYPES)

    return types


try:
    from PySide2 import QtWidgets
    import mincepy_gui

except ImportError:
    pass
else:

    class MinkipyActioner(mincepy_gui.Actioner):
        name = "minkipy-actioner"

        SUBMIT = 'Submit'

        def probe(self, obj: object, context: dict) -> Optional[Iterable[str]]:
            if self.is_task_or_task_record(obj) or self.is_collection_of_tasks(obj):
                return [self.SUBMIT]

            return None

        def do(self, action: str, obj: object, context: dict):
            to_submit = []

            def append(item):
                if isinstance(item, mincepy.DataRecord):
                    to_submit.append(mincepy.load(item.obj_id))
                elif isinstance(item, tasks.Task):
                    to_submit.append(item)

            if self.is_task_or_task_record(obj):
                append(obj)
            else:
                # Must be an iterable of tasks
                for entry in obj:
                    append(entry)

            if to_submit:
                queue_name = projects.working_on().default_queue

                parent = context[mincepy_gui.ActionContext.PARENT]
                queue_name, ok = QtWidgets.QInputDialog().getText(  # pylint: disable=invalid-name
                    parent,
                    "Submit task(s)",  # Title
                    "Queue name:",  # Label
                    echo=QtWidgets.QLineEdit.Normal,
                    text=queue_name)

                if ok:
                    if queue_name:
                        queues.queue(queue_name).submit(*to_submit)  # Submit the tasks
                    else:
                        QtWidgets.QMessageBox.warning(parent, "Submit cancelled",
                                                      "Must supply queue name")

        @staticmethod
        def is_task(obj: object) -> bool:
            return isinstance(obj, tasks.Task)

        @staticmethod
        def is_task_record(obj: object) -> bool:
            return isinstance(obj, mincepy.DataRecord) and obj.type_id == tasks.Task.TYPE_ID

        @staticmethod
        def is_task_or_task_record(entry: object) -> bool:
            return MinkipyActioner.is_task(entry) or MinkipyActioner.is_task_record(entry)

        @staticmethod
        def is_collection_of_tasks(obj: object) -> bool:
            # pylint: disable=isinstance-second-argument-not-valid-type
            return isinstance(obj, Iterable) and all(
                map(MinkipyActioner.is_task_or_task_record, obj))


def get_actioners():
    return (MinkipyActioner(),)



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = 'Martin Uhrin'
__license__ = 'GPLv3 and MIT, see LICENSE file'

about = {}
with open('minkipy/version.py') as f:
    exec(f.read(), about)

setup(name='minkipy',
      version=about['__version__'],
      description='Job and workflow submission made simple',
      long_description=open('README.rst').read(),
      url='https://github.com/muhrin/minkipy.git',
      author='Martin Uhrin',
      author_email='martin.uhrin.10@ucl.ac.uk',
      license=__license__,
      classifiers=[
          'License :: OSI Approved :: MIT License',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='workflows schedulers',
      install_requires=[
          'beautifultable~=1.0.0',
          'click',
          'mincepy>=0.15.15, <0.16',
          'kiwipy[rmq]~=0.6',
      ],
      extras_require={
          'dev': [
              'pip',
              'pytest>4',
              'pytest-cov',
              'pre-commit',
              'yapf',
              'prospector',
              'pylint',
              'twine',
          ],
          'docs': [
              'nbsphinx',
              'sphinx',
              'sphinx-autobuild',
          ],
          'pyos': ['pyos>=0.7.8', 'cmd2>=1.3.2'],
      },
      packages=[
          'minkipy',
          'minkipy.cli',
      ],
      include_package_data=True,
      test_suite='test',
      entry_points={
          'console_scripts': ['minki = minkipy.cli.main:minki'],
          'mincepy.plugins.types': ['minkipy_types = minkipy.provides:get_types',],
          'mincepy_gui.actioners': ['minkipy_gui_actioners = minkipy.provides:get_actioners',],
          'pyos.plugins.shell': ['pyos_commands = minkipy.pyos_extensions:get_commands',],
      })



# -*- coding: utf-8 -*-
import pathlib
import uuid

import mincepy
from mincepy import testing
import pytest  # pylint: disable=wrong-import-order

import minkipy

# pylint: disable=invalid-name


def simple_wf(obj):
    return obj


class Workflow(mincepy.SimpleSavable):
    TYPE_ID = uuid.UUID('ef45250f-b8b1-46bb-8033-01cd115483a6')

    def add_to_ten(self, value):  # pylint: disable=no-self-use
        return 10 + value


def test_python_command_from_method():
    # Static
    wf = Workflow()
    command = minkipy.command(wf.add_to_ten, args=(1,))  # type: minkipy.PythonCommand
    assert isinstance(command.script_file, mincepy.File)
    assert command.run() == 11

    # CHANGE:
    command = minkipy.command(wf.add_to_ten, args=(1,), dynamic=True)  # type: minkipy.PythonCommand
    assert isinstance(command.script_file, str)
    assert command.run() == 11


def test_python_command_from_module():
    from . import script
    command = minkipy.command(script, args=(3, 254))  # type: minkipy.PythonCommand
    assert isinstance(command.script_file, mincepy.File)
    assert command.run() == 257

    command = minkipy.command(script, args=(3, 254), dynamic=True)  # type: minkipy.PythonCommand
    assert isinstance(command.script_file, str)
    assert command.run() == 257

    # Now try using string instead of the module object
    # Just use some function that we know is importable
    command = minkipy.command('uuid@UUID', args=('61736206-729b-4a0b-9fac-6b5e71123ba0',))
    assert command.run().int == uuid.UUID('61736206-729b-4a0b-9fac-6b5e71123ba0').int


def test_saving_command_args():
    """Test that saving of arguments is handled correctly"""
    car = testing.Car()
    command = minkipy.command(simple_wf, args=(car,))

    command_id = command.save()
    car_id = car.obj_id

    assert command_id is not None
    assert car_id is not None

    del command

    command = mincepy.load(command_id)
    assert command.args[0].obj_id == car_id


def test_command_none_in_args():
    """This test checks for a bug where None could not be present in the args
    even though this should be perfectly possible"""
    command = minkipy.command(simple_wf, args=(None,))
    cid = command.save()
    del command

    loaded = mincepy.load(cid)  # type: minkipy.Command
    loaded.run()


def test_load_script_from_file():
    with minkipy.utils.working_directory(pathlib.Path(__file__).parent):
        cmd = minkipy.command('script.py@add', args=(5, 10))
        assert cmd.run() == 15

        # Try using the abspath now
        script_path = str(pathlib.Path('script.py').resolve())
        cmd = minkipy.command('{}@add'.format(script_path), args=(5, 10))
        assert cmd.run() == 15


def test_invalid_command():
    with pytest.raises(ValueError):
        minkipy.command('bla', type='made up')


def test_python_command_types():
    """Check that the various ways of creating PythonCommands work"""
    from . import script
    script_path = script.__file__

    # Non dynamic
    cmd = minkipy.PythonCommand(script_path, function='add', args=(5, 10))
    assert isinstance(cmd.script_file, mincepy.File)
    assert cmd.run() == 15

    # Dynamic
    cmd = minkipy.PythonCommand(script_path, function='add', args=(5, 10), dynamic=True)
    assert isinstance(cmd.script_file, str)
    assert cmd.run() == 15

    with pytest.raises(ValueError):
        minkipy.PythonCommand.build(5)


def test_command_saving_basics():
    """Test some basics of loading and saving a command"""
    # Check that older versions without 'dynamic' in their saved state can be loaded
    historian = mincepy.get_historian(create=False)

    from . import script
    script_path = script.__file__

    class CustomPythonCommand(minkipy.PythonCommand):

        def save_instance_state(self, saver):
            state = super().save_instance_state(saver)
            # Pretend we're an older version and have no kwargs or dynamic
            state.pop('_dynamic')
            state.pop('_kwargs')

    cmd = CustomPythonCommand(script_path, function='add', args=(5, 10))

    cmd_id = cmd.save()
    del cmd

    cmd = historian.load(cmd_id)
    assert cmd.kwargs == {}
    assert cmd.dynamic is False



# -*- coding: utf-8 -*-
import sys

import pprint
import click
import minkipy

pretty = pprint.PrettyPrinter(depth=4)  # pylint: disable=invalid-name


@click.group()
def minki():
    pass


@minki.command()
@click.option('--project', '-p', default=None, help='The project to use')
@click.option('--folder',
              '-f',
              default='',
              help='The folder to run the task in. '
              'Defaults to the id of the task as the folder name')
@click.option('--queue', '-q', default=None, help='The queue to send the task to')
@click.argument('cmd', type=str)
@click.argument('args', type=str, nargs=-1)
def submit(project, folder, queue, cmd, args):
    """Submit a task to a queue.  Will use the project default queue if not supplied."""
    proj = minkipy.workon(project)
    if queue is None:
        queue = proj.default_queue

    task = minkipy.task(cmd, args, folder)
    minkipy.queue(queue).submit(task)


@minki.command()
@click.option('--project', '-p', default=None, help='The project to use')
@click.option('--max-tasks', '-n', default=1, help='Number of tasks to process')
@click.option('--timeout',
              '-t',
              default=10.,
              help='The maximum time (in seconds) to wait for a new task')
@click.argument('queue', type=str, default=None, required=False)
def run(project, max_tasks, timeout, queue):
    """Process a number of tasks.  Will use the project default queue if not supplied."""
    proj = minkipy.workon(project)
    if queue is None:
        queue = proj.default_queue

    task_queue = minkipy.queue(queue)
    num_ran = minkipy.workers.run(task_queue, max_tasks, timeout)
    click.echo('Ran {} tasks'.format(num_ran))


@minki.command()
@click.option('--project', '-p', default=None, help='The project to use, defaults to active')
@click.option('--count', '-c', is_flag=True, help='Only show the number of tasks in each queue')
@click.argument('queues', type=str, nargs=-1)
def list(project, count: bool, queues):  # pylint: disable=redefined-builtin
    """List queued tasks.  Will use the project default queue if not supplied."""
    proj = minkipy.workon(project)
    if not queues:
        queues = (proj.default_queue,)

    for queue in queues:
        minki_queue = minkipy.queue(queue)
        click.echo('{}:'.format(queue))
        verbosity = 2 if not count else 0
        minki_queue.list(verbosity=verbosity)


@minki.command()
@click.option('--project', '-p', default=None, help='The project to use, defaults to active')
@click.argument('queue', type=str, default=None, required=False)
def purge(project, queue):
    """Remove all the tasks in a queue.  Will use the project default queue if not supplied."""
    proj = minkipy.workon(project)
    if queue is None:
        queue = proj.default_queue

    minki_queue = minkipy.queue(queue)
    num_purged = minki_queue.purge()
    click.echo('Cancelled {} tasks'.format(num_purged))


@minki.command()
@click.option('--project', '-p', default=None, help='The project to use, defaults to active')
@click.option('--queue', '-q', type=str, default=None, required=False)
@click.argument('tasks', type=str, nargs=-1)
def remove(project, queue, tasks):
    """Remove all the tasks in a queue.  Will use the project default queue if not supplied."""
    import mincepy

    proj = minkipy.workon(project)

    removed = []
    for task_id in tasks:
        if queue:
            this_queue = queue
            task = task_id
        else:
            # Try and get it from the task
            try:
                task = mincepy.load(task_id)  # type: minkipy.Task
            except mincepy.NotFound:
                click.echo("Task '{}' not found".format(task_id), err=True)
                continue
            else:
                this_queue = queue or task.queue
                if not this_queue:
                    this_queue = proj.default_queue

        if this_queue:
            this_queue = minkipy.queue(this_queue)
            this_removed = this_queue.remove(task)
            if this_removed:
                click.echo("Removed task '{}' from queue {}".format(task_id, this_queue.name))
                removed.extend(this_removed)
            else:
                click.echo('Failed to remove {}'.format(task_id), err=True)

    if len(removed) != len(tasks):
        sys.exit(1)

    return 0


@minki.command()
@click.argument('project', type=str)
def workon(project):
    """Set the active project"""
    try:
        minkipy.set_active_project(project)
    except ValueError as exc:
        click.echo(exc)
        return 1

    click.echo('Done')
    return 0
